#include	<stdio.h>

#define BLOCK_SIZE 	512

__inline__ __device__ int warpReduceSum ( int val ) 
{
	for ( int offset = warpSize/2; offset > 0; offset /= 2 )
		val += __shfl_down_sync (warpSize-1, val, offset );

	return val;
}

__inline__ __device__ int blockReduceSum ( int val ) 
{
	static __shared__ int shared [32];

	int lane = threadIdx.x % warpSize;
	int wid  = threadIdx.x / warpSize;

	val = warpReduceSum ( val );

		// write reduced value to shared memory
	if ( lane == 0 )
		shared [wid] = val;

	__syncthreads();

		// ensure we only grab a value from shared memory 
		// if that warp existed
	val = (threadIdx.x<blockDim.x/warpSize) ? shared[lane] : int(0);

	if ( wid == 0 )
		val = warpReduceSum ( val );

	return val;
}

__global__ void deviceReduceKernel ( int * in, int * out, int n ) 
{
	int sum = in [blockIdx.x * blockDim.x + threadIdx.x];
 
 	sum = blockReduceSum ( sum );

	if ( threadIdx.x == 0 )
		out [blockIdx.x] = sum;
}

__global__ void reduce5 ( int * inData, int * outData, int n )
{
	__shared__ int data [BLOCK_SIZE];
	int	  tid = threadIdx.x;
	int	  i   = blockIdx.x * blockDim.x + threadIdx.x;
	
	data [tid] = inData [i];		// load into shared memeory
	
	__syncthreads ();
	
	for ( int s = blockDim.x / 2; s > 32; s >>= 1 )
	{
		if ( tid < s )
			data [tid] += data [tid + s];
			
		__syncthreads ();
	}
	
	if ( tid < 32 )					// unroll last iterations
	{
		volatile int * ptr = data;

        	ptr [tid] += ptr [tid + 32]; 
	        ptr [tid] += ptr [tid + 16]; 
        	ptr [tid] += ptr [tid +  8]; 
	        ptr [tid] += ptr [tid +  4]; 
        	ptr [tid] += ptr [tid +  2]; 
	        ptr [tid] += ptr [tid +  1]; 
	}
	
	if ( tid == 0 )					// write result of block reduction
		outData [blockIdx.x] = data [0];
}

void testShuffle ( int * a, int * b, int n )
{
		// fill it with data
	for ( int i = 0; i < n; i++ )
		a [i] = i / 256;

		// perform reduce
	cudaEvent_t start, stop;
        float 	    gpuTime = 0.0f;

	cudaEventCreate ( &start );
	cudaEventCreate ( &stop  );
	cudaEventRecord ( start, 0 );

	int threads = 512;
	int blocks  = min((n + threads - 1) / threads, 1024);

	deviceReduceKernel<<<blocks, threads>>> ( a, b, n );
	
	cudaDeviceSynchronize ();

	cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	printf ( "Shuffle\n" );
	printf ( "time spent executing by the GPU: %.2f milliseconds\n", gpuTime );
}

void testReduce5 ( int * a, int * b, int n )
{
		// fill it with data
	for ( int i = 0; i < n; i++ )
		a [i] = i / 256;

		// perform reduce
	cudaEvent_t start, stop;
        float 	    gpuTime = 0.0f;

	cudaEventCreate ( &start );
	cudaEventCreate ( &stop  );
	cudaEventRecord ( start, 0 );

	int threads = 512;
	int blocks  = min((n + threads - 1) / threads, 1024);

	reduce5<<<blocks, threads>>> ( a, b, n );
	
	cudaDeviceSynchronize ();

	cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

	printf ( "Reduce5\n" );
	printf ( "time spent executing by the GPU: %.2f milliseconds\n", gpuTime );
}

int main ()
{
	int	n = 1024 * 1024 * 16;
	int   * a;
	int   * b;

		// allocate memory
	cudaMallocManaged ( &a, n * sizeof ( int ) );
	cudaMallocManaged ( &b, n * sizeof ( int ) / 512 );	// we need space for every block, ie n/512 elements

	testReduce5 ( a, b, n );
	testShuffle ( a, b, n );

	return 0;
}
