#include	<stdio.h>

__inline__ __device__ int warpReduceSum ( int val ) 
{
	for ( int offset = warpSize/2; offset > 0; offset /= 2 )
		val += __shfl_down_sync (warpSize-1, val, offset );

	return val;
}

__inline__ __device__ int blockReduceSum ( int val ) 
{
	static __shared__ int shared [32];
	int lane=threadIdx.x%warpSize;
	int wid=threadIdx.x/warpSize;

	val = warpReduceSum ( val );

		// write reduced value to shared memory
	if ( lane == 0 )
		shared [wid] = val;

	__syncthreads();

		// ensure we only grab a value from shared memory 
		// if that warp existed
	val = (threadIdx.x<blockDim.x/warpSize) ? shared[lane] : int(0);

	if ( wid == 0 )
		val = warpReduceSum ( val );

	return val;
}

__global__ void deviceReduceKernel ( int * in, int * out, int n ) 
{
	int sum = 0;

		//reduce multiple elements per thread
	for ( int i = blockIdx.x * blockDim.x + threadIdx.x; i < n; i += blockDim.x * gridDim.x ) 
		sum += in[i];
 
 	sum = blockReduceSum ( sum );

	if ( threadIdx.x == 0 )
		out [blockIdx.x] = sum;
}

void deviceReduce ( int * in, int * out, int n ) 
{
	int threads = 512;
	int blocks  = min((n + threads - 1) / threads, 1024);

	deviceReduceKernel<<<blocks, threads>>> ( in, out, n );
	deviceReduceKernel<<<1, 1024>>>         ( out, out, blocks );
}

int main ()
{
	int	n = 1024 * 1024 * 16;
	int   * a;
	int   * b;

		// allocate memory
	cudaMallocManaged ( &a, n * sizeof ( int ) );
	cudaMallocManaged ( &b, n * sizeof ( int ) / 512 );	// we need space for every block, ie n/512 elements

		// fill it with data
	for ( int i = 0; i < n; i++ )
		a [i] = i / 256;

		// perform reduce
	cudaEvent_t start, stop;
        float 	    gpuTime = 0.0f;

	cudaEventCreate ( &start );
	cudaEventCreate ( &stop  );
	cudaEventRecord ( start, 0 );

	deviceReduce ( a, b, n );

	cudaDeviceSynchronize ();

	cudaEventRecord      ( stop, 0 );
	cudaEventSynchronize ( stop );
	cudaEventElapsedTime ( &gpuTime, start, stop );

		// check the result
	int	sum = 0;

	for ( int i = 0; i < n; i++ )
		sum += a [i];

	printf ( "GPU %d, CPU %d\n", sum, b [0] );
	printf ( "time spent executing by the GPU: %.2f milliseconds\n", gpuTime );

	return 0;
}
